import 'dart:html' as html;
import 'dart:math';
import 'dart:typed_data';

import 'package:dio/dio.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';
import 'package:flutter_markdown/flutter_markdown.dart';
import 'package:google_fonts/google_fonts.dart';
import 'package:intl/intl.dart';
import 'package:remove_emoji/remove_emoji.dart';
import 'package:markdown/markdown.dart' as md;
import 'dart:ui' as ui;
import 'package:url_launcher/url_launcher.dart';
import 'package:dotted_border/dotted_border.dart';
import 'package:drop_zone/drop_zone.dart';
import 'package:dio/browser_imp.dart';
import 'package:dio/dio.dart' as Dio;
import 'package:flutter_spinkit/flutter_spinkit.dart';

abstract class Parser {
  Widget parse(Map block, {Options? options, BuildContext? context});
}

class Options {
  Options({
    this.disabled = false,
    this.loading = false,
    this.alignment,
    this.margin = const EdgeInsets.all(0),
    this.padding = const EdgeInsets.all(0),
    this.textAlign = TextAlign.center,
    this.textStyle = const TextStyle(
        fontFamily: "Raleway",
        fontSize: 14,
        fontWeight: FontWeight.w400,
        color: Colors.black),
    this.decoration = const BoxDecoration(),
  });

  bool disabled;
  bool loading;
  AlignmentGeometry? alignment;
  EdgeInsets margin;
  EdgeInsets padding;
  TextAlign textAlign;
  TextStyle textStyle;
  Decoration decoration;
}

class PlainTextWidget extends StatefulWidget {
  final Map? block;
  final Options? options;

  PlainTextWidget(this.block, {this.options});

  @override
  createState() => PlainTextWidgetState();
}

class PlainTextWidgetState extends State<PlainTextWidget> {
  @override
  Widget build(BuildContext context) {
    if (widget.block == null) return Container();

    var hasEmoji = widget.block!["emoji"];
    var innerType = widget.block!["type"];
    var innerText = widget.block!["text"];
    bool selectable = widget.block!["selectable"] ?? false;
    if (hasEmoji != null && hasEmoji == false) {
      innerText = RemoveEmoji().removemoji(innerText);
    }

    Widget innerWidget = Container();
    if (innerType == "plain_text") {
      innerWidget = selectable
          ? SelectableText(innerText,
              style: widget.options!.textStyle,
              textAlign: widget.options!.textAlign)
          : Text(innerText,
              style: widget.options!.textStyle,
              textAlign: widget.options!.textAlign);
    } else if (innerType == "mrkdwn") {
      innerWidget = MarkdownBody(
        selectable: selectable,
        data: innerText,
        extensionSet: md.ExtensionSet(
          md.ExtensionSet.gitHubFlavored.blockSyntaxes,
          [md.EmojiSyntax(), ...md.ExtensionSet.gitHubFlavored.inlineSyntaxes],
        ),
        styleSheet: MarkdownStyleSheet.fromTheme(Theme.of(context)).copyWith(
          p: widget.options!.textStyle.copyWith(height: 1.4),
        ),
        onTapLink: (text, href, title) {
          launch(href!);
        },
      );
    }

    return Container(
      alignment: widget.options!.alignment,
      padding: widget.options!.padding,
      decoration: widget.options!.decoration,
      child: innerWidget,
    );
  }
}

class PlainTextParser implements Parser {
  @override
  Widget parse(Map? block, {Options? options, BuildContext? context}) {
    return PlainTextWidget(block, options: options ?? Options());
  }
}

class ButtonWidget extends StatefulWidget {
  final Map block;
  final Options? options;

  ButtonWidget(this.block, {this.options});

  @override
  createState() => ButtonWidgetState();
}

class ButtonWidgetState extends State<ButtonWidget> {
  @override
  Widget build(BuildContext context) {
//var action = block["action"] != null ? block["action"] : {};
    var text = widget.block["text"];

    Widget inner = PlainTextParser().parse(text,
        context: context,
        options: Options(
            decoration: BoxDecoration(
                border: Border.all(width: 0.5, color: Colors.grey.shade600),
                borderRadius: BorderRadius.all(Radius.circular(4))),
            padding: EdgeInsets.fromLTRB(8, 6, 8, 6),
            textStyle: GoogleFonts.getFont('Open Sans',
                fontSize: 14,
                fontWeight: FontWeight.w600,
                color: Colors.black)));

    var loading = widget.options!.loading
        ? Container(
            margin: EdgeInsets.only(right: 4),
            child: SpinKitRing(
              color: Colors.amber,
              size: 18,
              lineWidth: 4,
            ),
          )
        : Container();

    return Container(
      margin: widget.options!.margin,
      child: Row(
        children: [
          loading,
          TextButton(
            onPressed: () {},
            child: inner,
            style: ButtonStyle(
                overlayColor: MaterialStateProperty.all(Colors.grey.shade100)),
          ),
        ],
      ),
    );
  }
}

class ButtonParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return ButtonWidget(block, options: options ?? Options());
  }
}

class SpacerParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return Spacer();
  }
}

class SectionParser implements Parser {
  const SectionParser();

  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    var tb = block["text"];
    if (tb != null) {
      return PlainTextParser().parse(tb,
          context: context,
          options: Options(
              textAlign: TextAlign.left,
              textStyle: GoogleFonts.getFont('Open Sans',
                  fontSize: 14,
                  fontWeight: FontWeight.w400,
                  color: Colors.black)));
    }
    return Container();
  }
}

class DividerParser implements Parser {
  const DividerParser();

  @override
  Widget parse(Map? block, {Options? options, BuildContext? context}) {
    return Divider(
      color: Colors.grey.shade500,
      height: 12,
      thickness: 0.5,
    );
  }
}

class HeaderParser implements Parser {
  const HeaderParser();

  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    var text = block["text"];
    return PlainTextParser().parse(
      text,
      context: context,
      options: Options(
        alignment: Alignment.centerLeft,
        padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
        textStyle: GoogleFonts.getFont('Open Sans',
            fontSize: 18, fontWeight: FontWeight.w700, color: Colors.black87),
      ),
    );
  }
}

class ImageWidget extends StatefulWidget {
  final Map? block;
  final Options? options;

  ImageWidget(this.block, {this.options});

  @override
  createState() => ImageWidgetState();
}

class ImageWidgetState extends State<ImageWidget> {
  Uint8List? _data = Uint8List.fromList([]);

  @override
  void initState() {
    init();
    super.initState();
  }

  Future<void> init() async {
    var url = widget.block!["image_url"];
    url = Uri.decodeFull(url);
    DioForBrowser dio = DioForBrowser();
    Response<dynamic> resp = await dio.request(
      url,
      options: Dio.Options(
        responseType: ResponseType.bytes,
      ),
    );

    setState(() {
      _data = resp.data;
    });
  }

  @override
  Widget build(BuildContext context) {
    var title = widget.block!["title"];

    Widget inner = PlainTextParser().parse(
      title,
      context: context,
      options: Options(
        alignment: Alignment.centerLeft,
        padding: EdgeInsets.fromLTRB(0, 8, 0, 4),
        textStyle: GoogleFonts.getFont(
          'Open Sans',
          fontSize: 14,
          fontWeight: FontWeight.w400,
          color: Colors.grey.shade700,
        ),
      ),
    );

    return Container(
        child: Column(
      children: [
        Row(
          children: [
            inner,
            Container(
              margin: EdgeInsets.fromLTRB(8, 4, 0, 0),
              child: SizedBox(
                width: 18,
                height: 18,
                child: TextButton(
                  onPressed: () {
                    init();
                  },
                  child: Container(child: Icon(Icons.refresh, size: 18)),
                  style: ButtonStyle(
                    foregroundColor: MaterialStateProperty.all(
                        widget.options!.disabled
                            ? Colors.grey.shade500
                            : Colors.amber.shade700),
                    backgroundColor:
                        MaterialStateProperty.all(Colors.transparent),
                    overlayColor: MaterialStateProperty.all(Colors.transparent),
                    padding: MaterialStateProperty.all(EdgeInsets.zero),
                  ),
                ),
              ),
            ),
          ],
        ),
        Container(
          alignment: Alignment.centerLeft,
          child: SizedBox(
            child: _data!.length == 0
                ? Container()
                : Image.memory(
                    _data!,
                    key: UniqueKey(),
                  ),
          ),
        ),
      ],
    ));
  }
}

class ImageParser implements Parser {
  const ImageParser();

  @override
  Widget parse(Map? block, {Options? options, BuildContext? context}) {
    return ImageWidget(block, options: options ?? Options());
  }
}

class DropWidget extends StatefulWidget {
  final Map block;
  final Options? options;
  final void Function(BuildContext context, Map<html.File, Uint8List>)?
      onDropCompleted;

  DropWidget(this.block, {this.onDropCompleted, this.options});

  @override
  createState() => DropWidgetState();
}

class DropWidgetState extends State<DropWidget> {
  Map<html.File, Uint8List?> files = {};
  bool reading = false;
  bool dragged = false;

  @override
  Widget build(BuildContext context) {
    bool disabled = files.length > 0 || widget.options!.disabled;
    var file = widget.block["file"] ?? {};
    String format = file["format"] ?? "";

    var loading = widget.options!.loading
        ? Container(
            margin: EdgeInsets.only(right: 16),
            child: SpinKitRing(
              color: Colors.amber,
              size: 18,
              lineWidth: 4,
            ),
          )
        : Container();

    return Row(
      children: [
        loading,
        GestureDetector(
          onTap: disabled
              ? null
              : () {
                  if (files.length > 0) return;

                  html.InputElement uploadInput =
                      html.FileUploadInputElement() as html.InputElement;
                  uploadInput.accept = format;
                  uploadInput.maxLength = file["max_count"] ?? 1;
                  uploadInput.click();

                  uploadInput.onChange.listen((e) {
                    var files = uploadInput.files!;
                    files = files
                        .where((element) => element.name.endsWith(format))
                        .toList()
                        .sublist(0, min(files.length, file["max_count"] ?? 1));

                    onDrop(files);
                    setState(() {
                      dragged = true;
                    });
                  });
                },
          child: Container(
            width: 450,
            height: 250,
            padding: EdgeInsets.all(12),
            color: dragged ? Colors.blueGrey.shade100 : Colors.grey.shade200,
            child: Stack(
              children: [
                DottedBorder(
                  color: disabled ? Colors.grey.shade500 : Colors.grey.shade800,
                  dashPattern: dragged ? [8, 0] : [8, 4],
                  strokeCap: StrokeCap.round,
                  strokeWidth: 1,
                  child: DropZone(
                    onDragEnter: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = true;
                            });
                          },
                    onDragExit: disabled
                        ? null
                        : () {
                            setState(() {
                              dragged = false;
                            });
                          },
                    onDrop: disabled
                        ? null
                        : (List<html.File>? files) {
                            var file = widget.block["file"] ?? {};
                            String format = file["format"] ?? "";

                            files = files!
                                .where(
                                    (element) => element.name.endsWith(format))
                                .toList()
                                .sublist(0,
                                    min(files.length, file["max_count"] ?? 1));

                            onDrop(files);
                          },
                    child: Container(
                      child: Center(
                        child: Column(
                          mainAxisAlignment: MainAxisAlignment.center,
                          children: [
                            Icon(
                              Icons.upload_file,
                              size: 72,
                              color: disabled
                                  ? Colors.grey.shade500
                                  : Colors.blueGrey,
                            ),
                            Container(
                              margin: EdgeInsets.only(top: 12, bottom: 12),
                              child: Row(
                                mainAxisAlignment: MainAxisAlignment.center,
                                children: getText(),
                              ),
                            )
                          ],
                        ),
                      ),
                    ),
                  ),
                ),
                Positioned(
                  right: 4,
                  top: 4,
                  child: SizedBox(
                    width: 32,
                    child: TextButton(
                      onPressed: widget.options!.disabled
                          ? null
                          : () {
                              setState(() {
                                files = {};
                                reading = false;
                                dragged = false;
                                widget.options!.loading = false;
                              });
                            },
                      child: Container(child: Icon(Icons.close)),
                      style: ButtonStyle(
                        foregroundColor: MaterialStateProperty.all(disabled
                            ? Colors.grey.shade500
                            : Colors.amber.shade700),
                        backgroundColor:
                            MaterialStateProperty.all(Colors.transparent),
                        overlayColor:
                            MaterialStateProperty.all(Colors.transparent),
                        padding: MaterialStateProperty.all(EdgeInsets.zero),
                      ),
                    ),
                  ),
                )
              ],
            ),
          ),
        ),
      ],
    );
  }

  List<Widget> getText() {
    bool disabled = files.length > 0 || widget.options!.disabled;

    if (reading) {
      return [
        Text("Preparing files...",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16,
                fontWeight: FontWeight.w700,
                color: Colors.blueGrey)),
      ];
    }

    if (files.length == 0) {
      var file = widget.block["file"] ?? {};
      String format = file["format"] ?? "";
      format = format.replaceAll(".", "").split(",").join(", ");
      return [
        Text("Choose a ${format ?? ""} file",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16,
                fontWeight: FontWeight.w700,
                color: disabled ? Colors.grey.shade500 : Colors.blueGrey)),
        Text(" or drag it here.",
            style: GoogleFonts.getFont('Open Sans',
                fontSize: 16,
                fontWeight: FontWeight.w400,
                color: disabled ? Colors.grey.shade500 : Colors.blueGrey))
      ];
    }

    List<Widget> names = [
      Text("Done!",
          style: GoogleFonts.getFont('Open Sans',
              fontSize: 16,
              fontWeight: FontWeight.w700,
              color: Colors.blueGrey))
    ];

    for (var file in files.keys) {
      names.add(Text(file.name,
          style: GoogleFonts.getFont('Open Sans',
              fontSize: 14,
              fontWeight: FontWeight.w700,
              color: Colors.blueGrey)));
    }

    return [
      Column(
        children: names,
      )
    ];
  }

  Future<void> onDrop(List<html.File> fileArr) async {
    if (fileArr.length == 0) {
      setState(() {
        dragged = false;
      });
      return;
    }

    setState(() {
      reading = true;
    });

    Map<html.File, Uint8List> _files = {};
    for (html.File file in fileArr) {
      html.FileReader reader = html.FileReader();
      reader.readAsArrayBuffer(file);
      await reader.onLoadEnd.first;
      _files[file] = reader.result as Uint8List;
    }

    if (widget.onDropCompleted != null)
      widget.onDropCompleted!(context, _files);

    setState(() {
      reading = false;
      files = _files;
      widget.options!.loading = true;
    });
  }
}

class DragDropParser implements Parser {
  Map<html.File, Uint8List> files = {};
  bool reading = false;

  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return DropWidget(
      block,
      onDropCompleted: (context, p1) => this.onDropCompleted(context, p1),
      options: options ?? Options(),
    );
  }

  void onDropCompleted(BuildContext context, Map<html.File, Uint8List> files) {}
}

class PlainTextInput extends StatefulWidget {
  final Map block;
  final Options? options;
  final void Function(Map) onCreated;
  final void Function(BuildContext, Map, String) onSubmitted;

  PlainTextInput(this.block, this.onCreated, this.onSubmitted, {this.options}) {
    this.onCreated(block);
  }

  @override
  createState() => PlainTextInputState();
}

class PlainTextInputState extends State<PlainTextInput> {
  final Map<TextInputType, List<TextInputFormatter>> formatters = {
    TextInputType.number: [
      FilteringTextInputFormatter.allow(RegExp(r'[0-9]')),
    ],
  };

  @override
  Widget build(BuildContext context) {
    var label = widget.block["label"];
    var loading = widget.options!.loading
        ? Container(
            margin: EdgeInsets.only(right: 8, bottom: 16),
            child: SpinKitRing(
              color: Colors.amber,
              size: 18,
              lineWidth: 4,
            ),
          )
        : Container();

    var vals = TextInputType.values.where((element) =>
        element
            .toJson()["name"]
            .toString()
            .replaceFirst("TextInputType.", "") ==
        widget.block["format"]);
    TextInputType inputType = vals.isNotEmpty ? vals.first : TextInputType.text;

    List<TextInputFormatter> formatter = formatters[inputType] ?? [];

    return Row(
      children: [
        loading,
        ConstrainedBox(
          constraints: BoxConstraints(
              minWidth: 240, maxWidth: 360, minHeight: 64, maxHeight: 64),
          child: Container(
            width: MediaQuery.of(context).size.width * 0.25,
            padding: EdgeInsets.only(left: 16),
            child: TextField(
              keyboardType: inputType,
              inputFormatters: formatter,
              enabled: !widget.options!.disabled,
              onSubmitted: (value) {
                widget.onSubmitted(context, widget.block, value);
              },
              decoration: InputDecoration(
                  border: OutlineInputBorder(),
                  labelText: label,
                  counterText: "\u23CE Press 'enter' to submit"),
            ),
          ),
        ),
      ],
    );
  }
}

class PlainTextInputParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return PlainTextInput(
      block,
      this.onCreated,
      this.onSubmitted,
      options: options ?? Options(),
    );
  }

  void onCreated(Map block) {}
  void onSubmitted(BuildContext context, Map block, String value) {}
}

class CheckboxWidget extends StatefulWidget {
  final Map block;
  final Options? options;
  final void Function(Map) onCreated;
  final void Function(BuildContext, Map, int, bool) onChanged;

  CheckboxWidget(this.block, this.onCreated, this.onChanged, {this.options}) {
    this.onCreated(block);
  }

  @override
  createState() => CheckboxWidgetState();
}

class CheckboxWidgetState extends State<CheckboxWidget> {
  List<bool?> values = [];

  @override
  Widget build(BuildContext context) {
    var options = widget.block["options"];
    List<Widget> checkboxes = [];

    int i = 0;
    for (var option in options) {
      var text = option["text"];
      var description = option["description"];
      bool? value = option["value"];
      values.add(value);

      var _i = i;
      checkboxes.add(
        IntrinsicWidth(
          child: CheckboxListTile(
            value: values[i],
            dense: true,
            controlAffinity: ListTileControlAffinity.leading,
            title: Column(
              crossAxisAlignment: CrossAxisAlignment.center,
              children: [
                PlainTextParser().parse(text,
                    context: context,
                    options: Options(
                        textStyle: GoogleFonts.getFont('Open Sans',
                            fontSize: 14,
                            fontWeight: FontWeight.w600,
                            color: Colors.grey.shade800))),
                PlainTextParser().parse(description,
                    context: context,
                    options: Options(
                        textStyle: GoogleFonts.getFont('Open Sans',
                            fontSize: 14,
                            fontWeight: FontWeight.w500,
                            color: Colors.grey.shade600))),
              ],
            ),
            onChanged: widget.options!.disabled
                ? null
                : (_val) {
                    setState(() {
                      values[_i] = _val;
                    });
                    widget.onChanged(context, widget.block, _i, _val ?? false);
                  },
          ),
        ),
      );

      i++;
    }

    return Container(
      padding: EdgeInsets.only(left: 16),
      child: IntrinsicWidth(
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: checkboxes,
        ),
      ),
    );
  }
}

class CheckboxParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return CheckboxWidget(block, this.onCreated, this.onChanged,
        options: options ?? Options());
  }

  void onCreated(Map block) {}
  void onChanged(BuildContext context, Map block, int index, bool value) {}
}

class RadioButtonWidget extends StatefulWidget {
  final Map block;
  final Options? options;
  final void Function(Map) onCreated;
  final void Function(BuildContext, Map, dynamic) onChanged;

  RadioButtonWidget(this.block, this.onCreated, this.onChanged,
      {this.options}) {
    this.onCreated(block);
  }

  @override
  createState() => RadioButtonWidgetState();
}

class RadioButtonWidgetState extends State<RadioButtonWidget> {
  String? value = "";

  @override
  void initState() {
    value = widget.block["value"] ?? widget.block["options"][0]["value"];
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    var options = widget.block["options"];
    List<Widget> radios = [];

    for (var option in options) {
      radios.add(
        IntrinsicWidth(
          child: RadioListTile(
            value: option["value"],
            groupValue: value,
            title: PlainTextParser().parse(option["text"],
                context: context,
                options: Options(
                    alignment: Alignment.centerLeft,
                    textStyle: GoogleFonts.getFont('Open Sans',
                        fontSize: 14, fontWeight: FontWeight.w500))),
            onChanged: widget.options!.disabled
                ? null
                : (dynamic _val) {
                    setState(() {
                      value = _val;
                    });
                    widget.onChanged(context, widget.block, _val);
                  },
          ),
        ),
      );
    }

    return Container(
      padding: EdgeInsets.only(left: 16),
      child: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        crossAxisAlignment: CrossAxisAlignment.center,
        children: radios,
      ),
    );
  }
}

class RadioButtonParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return RadioButtonWidget(block, this.onCreated, this.onChanged,
        options: options ?? Options());
  }

  void onCreated(Map block) {}
  void onChanged(BuildContext context, Map block, value) {}
}

class DateTimeWidget extends StatefulWidget {
  final Map block;
  final Options? options;
  final void Function(Map) onCreated;
  final void Function(BuildContext, Map, DateTime) onPicked;

  DateTimeWidget(this.block, this.onCreated, this.onPicked, {this.options}) {
    this.onCreated(block);
  }

  @override
  createState() => DateTimeWidgetState();
}

class DateTimeWidgetState extends State<DateTimeWidget> {
  DateTime selectedDate = DateTime.now();
  TextEditingController _controller = TextEditingController();
  final _f = new DateFormat("dd.MM.yyyy");

  @override
  void initState() {
    super.initState();
    if (widget.block["initial_date"] != null)
      selectedDate = DateTime.parse(widget.block["initial_date"]);
    _controller.text = _f.format(selectedDate);
  }

  @override
  Widget build(BuildContext context) {
    var loading = widget.options!.loading
        ? Container(
            margin: EdgeInsets.only(right: 4),
            child: SpinKitRing(
              color: Colors.amber,
              size: 18,
              lineWidth: 4,
            ),
          )
        : Container();

    return Row(
      children: [
        loading,
        ConstrainedBox(
          constraints: BoxConstraints(
              minWidth: 240, maxWidth: 240, minHeight: 64, maxHeight: 64),
          child: Container(
            margin: EdgeInsets.all(8),
            width: MediaQuery.of(context).size.width * 0.25,
            child: TextField(
              controller: _controller,
              style: GoogleFonts.getFont('Open Sans',
                  fontSize: 16,
                  fontWeight: FontWeight.w500,
                  color: widget.options!.disabled
                      ? Colors.grey.shade500
                      : Colors.grey.shade800),
              onTap: () async {
                DateTime? picked = await showDatePicker(
                    context: context,
                    initialDate: selectedDate,
                    firstDate: DateTime(1900),
                    lastDate: DateTime(2100));

                if (picked != null && picked != selectedDate) {
                  setState(() {
                    selectedDate = picked;
                  });
                  _controller.text = _f.format(selectedDate);
                  widget.onPicked(context, widget.block, picked);
                }
              },
              enabled: !widget.options!.disabled,
              readOnly: true,
              //onSubmitted: widget.onSubmitted,
              decoration: InputDecoration(
                border: OutlineInputBorder(),
                labelText: widget.block["placeholder"] != null
                    ? widget.block["placeholder"]["text"]
                    : "",
                suffixIcon: Container(
                  margin: EdgeInsets.only(right: 6),
                  //decoration: BoxDecoration(color: Colors.grey.shade200),
                  child: Icon(Icons.calendar_today),
                ),
              ),
            ),
          ),
        )
      ],
    );
  }
}

class DateTimePickerParser implements Parser {
  @override
  Widget parse(Map block, {Options? options, BuildContext? context}) {
    return DateTimeWidget(block, this.onCreated, this.onPicked,
        options: options ?? Options());
  }

  void onCreated(Map block) {}
  void onPicked(BuildContext context, Map block, DateTime date) {}
}
