library robomotion_app;

import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';
import 'dart:html' as html;

import 'package:dio/adapter_browser.dart';
import 'package:dio/browser_imp.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:dio/dio.dart';
import 'package:get/get.dart' as getx;
import 'package:google_fonts/google_fonts.dart';
import 'package:robomotion_app/encryption.dart';
import 'package:robomotion_app/errors.dart';
import 'package:robomotion_app/models/published_flow.dart';
import 'package:robomotion_app/models/queue_message.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import 'package:uuid/uuid.dart';

import 'package:encrypt/encrypt.dart' as enc;
import 'package:cryptography/cryptography.dart';
import 'package:convert/convert.dart';

StreamController<void> _keyExStreamCtrl = StreamController<void>.broadcast();
StreamController<void> _robotIDStreamCtrl = StreamController<void>.broadcast();

class RError {
  Map? data;
  RError() {
    this.data = null;
  }
}

class Robomotion {
  late RobomotionUI ui;
  final Config config;
  Function(QueueMessage data)? onMsgRecv;
  String? _apiURL;
  String? _appURL;
  String? _sid;
  late WebSocketChannel _channel;

  Robomotion({required this.config, this.onMsgRecv}) {
    _apiURL = _getAPIUrl();
    _appURL = _getAppUrl();
    _sid = _getCookie("sid") ?? "";
    generateAESKey();
    ui = RobomotionUI(lib: this);
  }

  Future<void> run() async {
    await _connectAppServer();
  }

  Future<List<PublishedFlow>> getPublishedFlows() async {
    String url = "${this._apiURL}/v1/flows.published.list";
    Map resp = await (_request(
            method: "GET",
            url: url,
            queryParams: {"page": 1, "size": 100, "search": ""})
        as FutureOr<Map<dynamic, dynamic>>);

    List<PublishedFlow> flows = [];
    for (dynamic f in resp["flows"]) {
      flows.add(PublishedFlow.fromJson(f));
    }
    return flows;
  }

  Future<String?> runFlow(String robotID, String flowID) async {
    String url = "${this._apiURL}/v1/apps.flow.run";
    String? _flowID = flowID;
    RError err = RError();

    final resp = await (_request(method: "POST", url: url, error: err, body: {
      "robot_id": robotID,
      "app_id": config._id,
      "flow_id": flowID
    }) as FutureOr<Map<dynamic, dynamic>>);

    if (err.data != null) {
      config.robotID = err.data!["robot_id"];
      try {
        var error = jsonDecode(err.data!["error"]);
        if (error.containsKey("payload") &&
            error["payload"].containsKey("flow_id")) {
          _flowID = error["payload"]["flow_id"];
        } else {
          throw Exception(error);
        }
      } catch (ex) {
        throw Exception(err.data!["error"]);
      }
    }

    config.robotID = resp.containsKey("robot_id") ? resp["robot_id"] : "";
    return _flowID;
  }

  Future<Map?> apiRequest({
    required String method,
    required String endpoint,
    Map? body,
    Map<String, dynamic>? queryParams,
    bool withCredentials = true,
    RError? error,
  }) async {
    if (!endpoint.startsWith("/")) {
      endpoint = "/$endpoint";
    }

    return _request(
      method: method,
      url: "$_apiURL$endpoint",
      body: body,
      queryParams: queryParams,
      withCredentials: withCredentials,
      error: error,
    );
  }

  Future<Map?> _request(
      {required String method,
      required url,
      Map? body,
      Map<String, dynamic>? queryParams,
      bool withCredentials = true,
      RError? error}) async {
    BrowserHttpClientAdapter adapter = BrowserHttpClientAdapter();
    adapter.withCredentials = withCredentials;

    DioForBrowser dio = DioForBrowser();
    dio.httpClientAdapter = adapter;

    Map<String, dynamic> headers = {
      'Content-Type': 'application/json; charset=UTF-8',
    };

    if (kDebugMode) {
      headers['Authorization'] = 'Bearer $_sid';
    }

    final Response<dynamic> resp = await dio.request(
      url,
      queryParameters: queryParams,
      data: body,
      options: Options(
        method: method,
        headers: headers,
      ),
    );

    if (!resp.data.containsKey("ok") || !resp.data["ok"]) {
      if (resp.data["error"] == ErrAuthFailed) {
        html.window.location.href = Uri.base.origin;
        return {};
      }
      if (error != null) {
        error.data = resp.data;
        return {};
      }
      throw new Exception(resp.data.toString());
    }

    return resp.data;
  }

  Future<void> _connectAppServer() async {
    String url = "$_appURL?id=${config._id}";

    if (kDebugMode) {
      //  url += "&sid=$_sid";
    }

    _channel = WebSocketChannel.connect(Uri.parse(url));
    _channel.stream.listen((data) {
      final msg = QueueMessage.fromJson(jsonDecode(data));
      if (msg.encrypted!) {
        msg.payload = decryptPayload(msg.payload!);
      }
      _onMsgRecv(msg).then((_) {
        if (onMsgRecv != null) onMsgRecv!(msg);
      });
    });
  }

  Future<void> _onMsgRecv(QueueMessage msg) async {
    if (msg.msgType == QMT_Error) {
      return;
    }

    if (config.robotID == "" && msg.msgType == QMT_RobotID) {
      await _robotIDStreamCtrl.stream.first;
    }

    switch (msg.msgType) {
      case QMT_AppRequest:
        final resp = QueueMessage(
            id: msg.id,
            appID: config._id,
            robotID: config.robotID,
            msgType: QMT_AppRequest,
            payload: utf8.encode('{"msg": "received"}') as Uint8List?,
            encrypted: true);

        sendAppMsg(resp);
        break;

      case QMT_KeyExchange:
        final pem = msg.payload!;
        robotPubKeyPem = utf8.decode(pem);
        robotPubKey = parsePublicKey(pem);
        _keyExStreamCtrl.add(null);
        break;

      case QMT_RobotID:
        config.robotID = utf8.decode(msg.payload!);
        _robotIDStreamCtrl.add(null);
        break;
    }
  }

  void sendAppMsg(QueueMessage msg) {
    if (msg.encrypted!) {
      msg.payload = encryptPayload(msg.payload!);
    }
    _channel.sink.add(msg.toString());
  }

  Future<void> keyExchange() async {
    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: config._id,
      robotID: config.robotID,
      userID: config._userID,
      msgType: QMT_KeyExchange,
      payload: null,
      encrypted: false,
    );

    sendAppMsg(msg);
    await _keyExStreamCtrl.stream.first;

    final encAes = rsaEncrypt(robotPubKey, aesKey.bytes);
    msg = QueueMessage(
        id: Uuid().v4(),
        appID: config._id,
        robotID: config.robotID,
        msgType: QMT_KeyExchange,
        payload: encAes,
        encrypted: false);

    sendAppMsg(msg);
  }

  String? appIn(Map payload, {appID, robotID}) {
    var msg = QueueMessage(
      id: Uuid().v4(),
      appID: appID == null ? config._id : appID,
      robotID: robotID == null ? config.robotID : robotID,
      userID: config._userID,
      msgType: QMT_AppIn,
      payload: utf8.encode(jsonEncode({"payload": payload})) as Uint8List?,
      encrypted: true,
    );

    sendAppMsg(msg);
    return msg.id;
  }

  Future<String> _getUploadLink(String file) async {
    String url = "${this._apiURL}/v1/apps.upload.get";
    var resp = await (_request(
            method: "GET",
            url: url,
            queryParams: {"app_id": config._id, "file": file})
        as FutureOr<Map<dynamic, dynamic>>);
    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }
    return Uri.decodeFull(resp["url"]);
  }

  Future<void> uploadFile(String file, Uint8List data) async {
    String url = await _getUploadLink(file);
    data = encryptPayload(data);
    await DioForBrowser().put(url,
        data: base64Encode(data),
        options:
            Options(headers: {"Content-Type": "application/octet-stream"}));
  }

  Future<String?> getFileURL(String file) async {
    String url = "${this._apiURL}/v1/apps.file.get";
    var resp = await (_request(
            method: "GET",
            url: url,
            queryParams: {"app_id": config._id, "file": file})
        as FutureOr<Map<dynamic, dynamic>>);
    if (!resp.containsKey("url")) {
      throw Exception(resp.toString());
    }
    return resp["url"];
  }

  String _getAPIUrl() {
    String ws = this.config.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0];
    return domain == "robomotion.io" ? "$protocol//api.$domain" : "$ws/api";
  }

  String _getAppUrl() {
    String ws = this.config.workspace ?? Uri.base.origin;
    List<String> parts = ws.split(".");
    String domain = parts.getRange(1, parts.length).join(".");
    String protocol = parts[0].split("//")[0] == "https:" ? "wss:" : "ws:";
    return domain == "robomotion.io"
        ? "$protocol//mq.${domain}/app"
        : "$protocol//${parts[0].split("//")[1]}.$domain/mq/app";
  }

  Future<Map?> getUser() async {
    var resp = await apiRequest(
      method: 'GET',
      endpoint: '/v1/users.get',
    );
    if (resp == null) {
      return null;
    }
    if (!resp['ok']) {
      return null;
    }

    return resp['user'];
  }

  Future<String> getPrivateKey(Map user, String password) async {
    final encPrivateKey = user['enc_private_key'];
    final salt = user['salt'];
    final iv = user['iv'];
    final username = user['email'];

    final algorithm = Hkdf(hmac: Hmac(Sha256()), outputLength: 32);
    final derivedSalt = await algorithm.deriveKey(
      secretKey: SecretKey(utf8.encode(username)),
      nonce: hex.decode(salt),
    );

    final pbkdf2 = Pbkdf2(
      macAlgorithm: Hmac.sha256(),
      iterations: 100000,
      bits: 256,
    );

    final mk = await pbkdf2.deriveKey(
      secretKey: SecretKey(utf8.encode(password)),
      nonce: utf8.encode(hex.encode(await derivedSalt.extractBytes())),
    );

    final _encrypter = enc.Encrypter(
      enc.AES(
        enc.Key(
          Uint8List.fromList(await mk.extractBytes()),
        ),
        mode: enc.AESMode.cbc,
      ),
    );

    final pkPem = _encrypter.decryptBytes(
      enc.Encrypted(Uint8List.fromList(hex.decode(encPrivateKey))),
      iv: enc.IV(Uint8List.fromList(hex.decode(iv))),
    );

    return utf8.decode(pkPem);
  }
}

class Config {
  String _id = "";
  //final String namespace;
  //final String token;
  String _userID = "";
  //final String repositoryID;
  final String? workspace;
  String? robotID;
  String? flowID;
  late Map _profile;

  Config(
      { //@required this.id,
      //@required this.namespace,
      //@required this.token,
      //@required this.userID,
      //@required this.repositoryID,
      this.workspace}) {
    if (kDebugMode) {
      this._id = _getCookie("app_id") ?? "";
    } else {
      List<String> parts = Uri.base.toString().split("/");
      for (String p in parts.reversed) {
        try {
          Uuid.parse(p);
          this._id = p;
          break;
        } catch (e) {}
      }
    }

    var cookie = _getCookie("profile");
    if (cookie != null) {
      this._profile =
          jsonDecode(Uri.decodeFull(cookie)) ?? "{}" as Map<dynamic, dynamic>;
      this._userID = this._profile["id"] ?? "";
    }
  }
}

String? _getCookie(String cname) {
  if (html.document.cookie == "") return null;

  var name = cname + "=";
  var ca = html.document.cookie!.split(';');

  for (var i = 0; i < ca.length; i++) {
    var c = ca[i];
    while (c[0] == ' ') {
      c = c.substring(1);
    }
    if (c.indexOf(name) == 0) {
      return c.substring(name.length, c.length);
    }
  }

  return null;
}

class RobomotionUI extends StatefulWidget {
  final Robomotion lib;
  RobomotionUI({Key? key, required this.lib}) : super(key: key);

  @override
  _RobomotionUIState createState() => _RobomotionUIState();

  Future<String> askPassword(BuildContext context) async =>
      createState().askPassword(context);

  void snackbar(
    String title,
    String message, {
    Color backgroundColor = Colors.green,
  }) =>
      createState().snackbar(
        title,
        message,
        backgroundColor: backgroundColor,
      );
}

class _RobomotionUIState extends State<RobomotionUI> {
  String password = "";
  @override
  Widget build(BuildContext context) {
    return Container();
  }

  void snackbar(
    String title,
    String message, {
    Color backgroundColor = Colors.green,
  }) {
    getx.Get.snackbar(
      title,
      message,
      snackPosition: getx.SnackPosition.BOTTOM,
      backgroundColor: backgroundColor,
      colorText: Colors.white,
      maxWidth: 720,
      onTap: (snack) {
        getx.Get.closeCurrentSnackbar();
      },
    );
  }

  Future<String> askPassword(BuildContext context) async {
    int resp = await showDialog<int>(
          context: context,
          builder: (BuildContext context) => AlertDialog(
            title: const Text('Welcome Back!'),
            content: Column(
              mainAxisAlignment: MainAxisAlignment.start,
              mainAxisSize: MainAxisSize.min,
              children: [
                Row(
                  children: [
                    Text(
                      'Password',
                      style: GoogleFonts.openSans(
                        fontSize: 14,
                        fontWeight: FontWeight.w600,
                      ),
                    ),
                  ],
                ),
                SizedBox(
                  width: 240,
                  child: TextFormField(
                    obscureText: true,
                    enableSuggestions: false,
                    autocorrect: false,
                    style: GoogleFonts.openSans(
                      fontSize: 14,
                      fontWeight: FontWeight.w500,
                    ),
                    decoration: const InputDecoration(
                      icon: Icon(
                        Icons.password,
                      ),
                      enabledBorder: UnderlineInputBorder(
                        borderSide: BorderSide(),
                      ),
                      hintText: 'Your Password',
                    ),
                    onChanged: (text) {
                      setState(() {
                        password = text;
                      });
                    },
                  ),
                ),
              ],
            ),
            actions: [
              Builder(
                builder: (context) {
                  bool enabled = password.isNotEmpty;
                  return ElevatedButton(
                    style: ButtonStyle(
                      padding: MaterialStateProperty.all(
                        const EdgeInsets.all(16),
                      ),
                      backgroundColor: MaterialStateProperty.all(
                        enabled ? Colors.blue : Colors.grey,
                      ),
                      shape: MaterialStateProperty.all(
                        RoundedRectangleBorder(
                          borderRadius: BorderRadius.circular(8),
                        ),
                      ),
                    ),
                    onPressed: enabled
                        ? () {
                            Navigator.pop(context, 0);
                          }
                        : null,
                    child: Text(
                      'UNLOCK',
                      style: GoogleFonts.openSans(
                        fontSize: 14,
                        fontWeight: FontWeight.w600,
                        color: Colors.white,
                      ),
                    ),
                  );
                },
              ),
            ],
          ),
        ) ??
        -1;

    if (password.isEmpty) return '';

    try {
      Map user = await widget.lib.getUser() ?? {};
      await widget.lib.getPrivateKey(user, password);
    } catch (e) {
      snackbar(
        'Invalid Password',
        e.toString(),
        backgroundColor: Colors.red,
      );
      return '';
    }

    return password;
  }
}
